/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2006.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * ----------------------------------------------------------------------------
 *
 *  AUTHOR:      Dirk Tepelmann
 *  DATE:        June, 2005
 *
 *  REVISION INFO:
 *        $Revision$ $Date$
 * ----------------------------------------------------------------------------- */
package com.testingtech.ttcn.tci.codec.helper;
import org.etsi.ttcn.tci.IntegerValue;
import org.etsi.ttcn.tci.RecordOfValue;
import org.etsi.ttcn.tci.RecordValue;
import org.etsi.ttcn.tci.UnionValue;

import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularEspHeader;
import com.testingtech.ttcn.tci.codec.esp.EspHeaderCodec;
import com.testingtech.ttcn.tci.codec.tabular.TabularDecoder;
import com.testingtech.ttcn.tci.codec.tabular.TabularException;
import com.testingtech.ttcn.tci.codec.tabular.TabularObject;
import com.testingtech.util.BitArrayInputStream;

/**
 * @author dte
 *
 */
public class SupportFunctions {

	private final static int HOP = 0;
	private final static int TUNNELEDIPV6 = 41;
	private final static int ROUTING = 43;
	private final static int FRAGMENT = 44;
	private final static int ESP = 50;
	private final static int AUTHENTICATION = 51;
	private final static int DESTINATION = 60;
	private final static int MOBILITY = 135;

	private static boolean isESP = false;
	private static RecordValue espHeader = null;
	
	/**
	 * Only for ICMP.
	 * @param dec
	 * @param header
	 * @param type
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkHeaderAndType(TabularDecoder dec, int header, int type) throws TabularException {
		Integer pos = 0;
		if ((pos=innerCheckHeader(dec, header)) != -1) {
			if (!isESP) {
				if (dec.bytes(pos, 1, type)) {
					return true;
				}
			}
			else {
				if (espHeader != null) {
					RecordValue espPayload;
					RecordValue espIpDatagram;
					UnionValue ipv6Payload;
					espPayload = (RecordValue)espHeader.getField("espPayload");
					espIpDatagram = (RecordValue)(espPayload).getField("espIpDatagram");
					if (!espIpDatagram.notPresent()) {
						ipv6Payload = (UnionValue)espIpDatagram.getField("ipv6Payload");
						if (!ipv6Payload.notPresent()) {
							//we can ensure that we will handle an ICMP payload(nextHeader was checked before)
							RecordValue variant = (RecordValue)ipv6Payload.getVariant(ipv6Payload.getPresentVariantName());
							int icmpType = ((IntegerValue)variant.getField("icmpType")).getInt();
							if (icmpType == type) {
								return true;
							}
						}
					}
				}
				else {
					throw new TabularException("checkHeaderAndType: Cannot evaluate ESPHeader!");
				}
			}
				
		}
		
		return false;
	}

	/**
	 * @param dec
	 * @param header
	 * @return
	 * @throws TabularException
	 */
	private static int innerCheckHeader(TabularDecoder dec, int header) throws TabularException {
		int pos=6;
		int nextHeader = dec.getBytes(pos, 1);
		
		isESP = false;
		espHeader = null;
		
		//point to the end of the IPv6Header
		pos+= 34;
		//we do not want to go further in case of an mobility header,
		//because no further payload is expected
		while (isExtensionHeader(nextHeader) && (nextHeader != MOBILITY)) {
			switch (nextHeader) {
			case TUNNELEDIPV6:
				nextHeader = dec.getBytes(pos+6, 1);
				pos+=40;
				break;
			case FRAGMENT:
				nextHeader = dec.getBytes(pos, 1);
				pos+=8;
				break;
			case AUTHENTICATION:
				nextHeader = dec.getBytes(pos, 1);
				pos+= ((dec.getBytes(pos+1, 1) + 2) * 4);
				break;
			case ESP:
				if ((nextHeader=handleESPHeader(dec, pos))==header) {
					//just do indicate that the header is correct
					return 1;
				}
				break;
			default:
				nextHeader = dec.getBytes(pos, 1);
				pos += (dec.getBytes(pos+1, 1) * 8) + 8;
				break;
			}
		}

		return (nextHeader == header)?pos:-1;
	}

	/**
	 * Decrypts an encrypted packet to retrieve the nextheader value and returns this value.
	 * @param dec
	 * @param pos
	 * @return
	 * @throws TabularException
	 */
	private static int handleESPHeader(TabularDecoder dec, int pos) throws TabularException {
		int amount = dec.infinity()/8-pos;
		byte [] ba = new byte [amount];
		for (int i=0; i<amount; i++) {
			ba[i] = (byte)dec.getBytes(pos+i, 1);
		}
		isESP = true;
		TabularDecoder decESP = new TabularDecoder(new BitArrayInputStream(ba), dec.getRb());
		decESP.setTagMap(dec.getTagMap());
		decESP.intTag("IPv6Header_payloadLength", ba.length);
        decESP.intTag("hdrExtIntLen", dec.getIntTag("IPv6Header_payloadLength"));
		TabularEspHeader tEspHeader = new TabularEspHeader(dec.getRb().tciCDRequired
				.getTypeForName("LibIpv6_Interface_TypesAndValues.EspHeader")
				.newInstance());
		EspHeaderCodec.decodeESPHeader(decESP, (TabularObject)tEspHeader);
		espHeader = tEspHeader.rv;
		return getNextHeaderFromESPHeader(espHeader, decESP, 0);
	}
	
	/**
	 * Retrieves the nextheader value from an encrypted packet.
	 * @param rv
	 * @param dec
	 * @param pos
	 * @return
	 * @throws TabularException
	 */
	private static int getNextHeaderFromESPHeader(RecordValue rv, TabularDecoder dec, int pos) throws TabularException {
		int nextHeader = ((IntegerValue)rv.getField("nextHeader")).getInt();
		if (isExtensionHeader(nextHeader)) {
			RecordValue espPayload = (RecordValue)rv.getField("espPayload");
			RecordValue espIpDatagram = (RecordValue)espPayload.getField("espIpDatagram");
			UnionValue extHdr = null;
			if (!espIpDatagram.notPresent()) {
				RecordOfValue extHdrList = (RecordOfValue)espIpDatagram.getField("extHdrList");
				if (!extHdrList.notPresent()) {
					extHdr = (UnionValue)extHdrList.getField(extHdrList.getLength()-1);
					if (extHdr.getPresentVariantName().equals("espHeader")) {
						nextHeader = handleESPHeader(dec, pos);
					}
					else if (extHdr.getPresentVariantName().equals("mobileHeader")){
//						nextHeader = ((IntegerValue)((RecordValue)extHdr.getVariant(extHdr.getPresentVariantName())).getField("payloadProtocol")).getInt();
						//payload protocol should always indicate 59(none)
						//so we decide that the last next header value is mobility
						nextHeader = 135;
					}
					else {
						nextHeader = ((IntegerValue)((RecordValue)extHdr.getVariant(extHdr.getPresentVariantName())).getField("nextHeader")).getInt();
					}
//					if (extHdr.getType().getName().equals("EspHeader")) {
//						nextHeader = handleESPHeader(dec, pos);
//					}
//					else if (extHdr.getType().getName().equals("MobileHeader")) {
//						nextHeader = ((IntegerValue)rv.getField("payloadProtocol")).getInt();
//					}
//					else {
//						nextHeader = ((IntegerValue)rv.getField("nextHdr")).getInt();
//					}
				}
			}
		}
		return nextHeader;
	}
	
	/**
	 * Checks if nextheader is an extension header.
	 * @param nextHeader
	 * @return
	 */
	public static boolean isExtensionHeader(int nextHeader) {
		return (nextHeader == HOP) || (nextHeader == TUNNELEDIPV6)
				|| (nextHeader == ROUTING) || (nextHeader == FRAGMENT)
				|| (nextHeader == DESTINATION) || (nextHeader == ESP)
				|| (nextHeader == AUTHENTICATION) || (nextHeader == MOBILITY);
	}
	
	/**
	 * @param dec
	 * @param header
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkHeader(TabularDecoder dec, int header) throws TabularException {
		return (innerCheckHeader(dec, header)!=-1);
	}

	/**
	 * @param dec
	 * @param type
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkMobileHeaderAndType(TabularDecoder dec, int type) throws TabularException {
		Integer pos = 0;
		if ((pos=innerCheckHeader(dec, 135)) != -1) {
			if (!isESP) {
				// +2 to get the mobility header type
				if (dec.bytes(pos+2, 1, type)) {
					return true;
				}
			}
			else {
				if (espHeader != null) {
					RecordValue espPayload;
					RecordValue espIpDatagram;
					RecordOfValue extHdrList;
					espPayload = (RecordValue)espHeader.getField("espPayload");
					espIpDatagram = (RecordValue)(espPayload).getField("espIpDatagram");
					if (!espIpDatagram.notPresent()) {
						extHdrList = (RecordOfValue)espIpDatagram.getField("extHdrList");
						if (!extHdrList.notPresent()) {
							for (int i=0; i<extHdrList.getLength(); i++) {
								UnionValue extHdr = (UnionValue)extHdrList.getField(i);
								if (extHdr.getPresentVariantName().equals("mobileHeader")) {
									RecordValue mobileHeader = (RecordValue)extHdr.getVariant(extHdr.getPresentVariantName());
									int mobileHeaderType = ((IntegerValue)mobileHeader.getField("mobileHeaderType")).getInt();
									if (mobileHeaderType == type) {
										return true;
									}
								}
							}
						}
					}
				}
				else {
					throw new TabularException("checkMobileHeaderAndType: Cannot evaluate ESPHeader!");
				}
			}
		}
		return false;
	}

	/**
	 * @param dec
	 * @param header
	 * @param type
	 * @param subtype
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkHeaderAndTypeAndSubtype(TabularDecoder dec, int header, int type, int subtype) throws TabularException {
		Integer pos = 0;
		if ((pos=innerCheckHeader(dec, header)) != -1) {
			if (!isESP) {
				if (dec.bytes(pos, 1, type)) {
					if (dec.bytes(pos+4, 1, subtype)) {
						return true;
					}
				}
			}
			else {
				if (espHeader != null) {
					RecordValue espPayload;
					RecordValue espIpDatagram;
					UnionValue ipv6Payload;
					espPayload = (RecordValue)espHeader.getField("espPayload");
					espIpDatagram = (RecordValue)(espPayload).getField("espIpDatagram");
					if (!espIpDatagram.notPresent()) {
						ipv6Payload = (UnionValue)espIpDatagram.getField("ipv6Payload");
						if (!ipv6Payload.notPresent()) {
							//we can ensure that we will handle an ICMP payload(nextHeader was checked before)
							RecordValue variant = (RecordValue)ipv6Payload.getVariant(ipv6Payload.getPresentVariantName());
							int icmpType = ((IntegerValue)variant.getField("icmpType")).getInt();
							if (icmpType == type) {
								int subType = ((IntegerValue)variant.getField("subtype")).getInt();
								if (subType == subtype) {
									return true;
								}
							}
						}
					}
				}
				else {
					throw new TabularException("checkHeaderAndTypeAndSubtype: Cannot evaluate ESPHeader!");
				}
			}
		}
		return false;
	}

	/**
	 * Checks that the UDP payload is really an IKE message.
	 * @param dec
	 * @return
	 * @throws TabularException
	 */
	public static boolean isIKEmsg(TabularDecoder dec) throws TabularException {
		Integer pos = 0;
		if ((dec.bytes(pos, 2, 500) || dec.bytes(pos, 2, 4500) || dec.bytes(pos, 2, 501)) //501 added for TC_SEC_6206_01
				&& (dec.bytes(pos + 2, 2, 500) || dec.bytes(pos + 2, 2, 4500) || dec.bytes(pos + 2, 2, 501))) {
			return true;
		}
		return false;
	}

	/**
	 * @param dec
	 * @param type
	 * @param flag
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkIKEmsg(TabularDecoder dec, int type, int flag) throws TabularException {
		Integer pos = 0;
		int srcPort = 0;
		if ((pos=innerCheckHeader(dec, 17)) != -1) {
			if (!isESP) {
				srcPort = dec.getBytes(pos, 2);
				if (!((dec.bytes(pos, 2, 500) || dec.bytes(pos, 2, 4500) || dec.bytes(pos, 2, 501)) //501 added for TC_SEC_6206_01
						&& (dec.bytes(pos + 2, 2, 500) || dec.bytes(pos + 2, 2, 4500) || dec.bytes(pos + 2, 2, 501)))) {
					return false;
				}
			}
			else {
				if (espHeader != null) {
					RecordValue espPayload;
					RecordValue espIpDatagram;
					UnionValue ipv6Payload;
					espPayload = (RecordValue)espHeader.getField("espPayload");
					espIpDatagram = (RecordValue)(espPayload).getField("espIpDatagram");
					if (!espIpDatagram.notPresent()) {
						ipv6Payload = (UnionValue)espIpDatagram.getField("ipv6Payload");
						if (!ipv6Payload.notPresent()) {
							//we can ensure that we will handle an UDP payload(nextHeader was checked before)
							RecordValue variant = (RecordValue)ipv6Payload.getVariant(ipv6Payload.getPresentVariantName());
							int sourcePort = ((IntegerValue)variant.getField("sourcePort")).getInt();
							int destPort = ((IntegerValue)variant.getField("destPort")).getInt();
							if (!( ((sourcePort == 500) || (sourcePort == 4500) || (sourcePort == 501)) //501 added for TC_SEC_6206_01
									&& ((destPort == 500) || (destPort == 4500) || (destPort == 501)))){
								return false;
							}
							int exchangeType = ((IntegerValue)variant.getField("exchangeType")).getInt();
							if (exchangeType != type) {
								return false;
							}
							int rFlag = ((IntegerValue)variant.getField("rFlag")).getInt();
							if (rFlag == flag) {
								return true;
							}
						}
					}
				}
				else {
					throw new TabularException("checkIKEmsg: Cannot evaluate ESPHeader!");
				}
			}
		}
		//add padding to position if IKE port is 4500
		if (srcPort == 4500) {
			pos += 4;
		}
		if (!dec.bytes(pos+26, 1, type)) {
			return false;
		}
		if (dec.bits((pos+27)*8+2, 1, flag)) {
			return true;
		}
		
		return false;
	}

	/**
	 * Only for ICMP.
	 * @param dec
	 * @param protocol
	 * @param type
	 * @return
	 * @throws TabularException
	 */
	public static boolean checkIpv4ICMPAndType(TabularDecoder dec, int protocol, int type) throws TabularException {

		if (dec.getBytes(9, 1) != protocol) {
			return false;
		}

		return dec.bytes((dec.getBytes(0, 1)&0x0F)*4, 1, type);
	}

}
